from django.contrib import admin

from cache.models import Invalidation


class InvalidationAdmin(admin.ModelAdmin):
    list_display = ('key', 'object_id', 'count', 'created_at')
    search_fields = ('key',)


admin.site.register(Invalidation, InvalidationAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Invalidation',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('key', models.CharField(help_text=b'cache key', max_length=255, db_index=True)),
                ('object_id', models.IntegerField()),
                ('sql', models.TextField(null=True)),
                ('count', models.IntegerField(null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('class_name', models.CharField(max_length=255, null=True)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType', null=True)),
            ],
        ),
        migrations.AlterIndexTogether(
            name='invalidation',
            index_together=set([('content_type', 'object_id')]),
        ),
    ]



# coding: utf-8
import hashlib
import logging

from django.conf import settings
from django.core.cache import caches
from django.db import models
from django.db.models.query import QuerySet
from django.db.models.sql import EmptyResultSet

from cache.constants import ONE_DAY
from cache.service import CacheService

cache = caches['test'] if settings.TESTING else caches['cache_manager']
logger = logging.getLogger(__name__)


class CachedQuerySet(QuerySet):
    def __init__(self, *args, **kwargs):
        self.timeout = kwargs.get('timeout', ONE_DAY)
        if 'timeout' in kwargs:
            del kwargs['timeout']
        super(CachedQuerySet, self).__init__(*args, **kwargs)

    def get_cache_key_and_sql(self):
        sql = self._sql()
        model = self.model._meta.db_table
        key = u'{namespace}::{model}::{sql}::{db}'.format(
            namespace=CacheService.get_namespace(model),
            model=model,
            sql=sql,
            db=self.db,
        )
        return hashlib.md5(key.encode('utf-8')).hexdigest(), sql

    def _sql(self):
        clone = self.query.clone()
        sql, params = clone.get_compiler(using=self.db).as_sql()
        return sql % params

    def iterator(self):
        try:
            key, sql = self.get_cache_key_and_sql()
        # workaround for Django bug # 12717
        except EmptyResultSet:
            return

        cached_objects = cache.get(key)
        if not cached_objects:
            logger.debug('cache miss for {}'.format(key))
            cached_objects = list(super(CachedQuerySet, self).iterator())

            # be careful, if the cache value is too large and your memcached
            # is running on small memory box, the key will be evicted immediately
            # after your inserted it. So that will cause legacy invalidation creation
            # in the invalidation table
            CacheService.bulk_create_invalidation(cached_objects, key, sql)
            cache.set(key, cached_objects, timeout=self.timeout)
        else:
            logger.debug('cache hit for {}'.format(key))

        for obj in cached_objects:
            yield obj

    def bulk_create(self, *args, **kwargs):
        # TODO invalidate cached objects
        return super(CachedQuerySet, self).bulk_create(*args, **kwargs)

    def update(self, **kwargs):
        # TODO invalidate cached objects
        return super(CachedQuerySet, self).update(**kwargs)


class CacheManager(models.Manager):
    def __init__(self, *args, **kwargs):
        self.timeout = kwargs.get('timeout', ONE_DAY)
        if 'timeout' in kwargs:
            del kwargs['timeout']
        super(CacheManager, self).__init__(*args, **kwargs)

    def get_queryset(self):
        return CachedQuerySet(self.model, using=self._db, timeout=self.timeout)



# coding:utf-8
import logging

logger = logging.getLogger(__name__)


def invalidate_model_cache(sender, instance, **kwargs):
    from cache.service import CacheService

    # TODO there is an assumption that we should use cached_objects
    # instead of objects. but there is not limit for this in code.
    # so if user use objects for CacheManager, the invalidation logic
    # is not correct.
    if not hasattr(instance.__class__, 'cached_objects'):
        return
    logger.debug(u'Invalidating model cache for {} {}'.format(sender, instance))
    CacheService.invalidate_cache_for(instance)


def invalidate_m2m_cache(sender, instance, model, **kwargs):
    from cache.service import CacheService

    if not hasattr(instance.__class__, 'cached_objects'):
        return
    logger.debug(u'Invalidating m2m cache for {} {} {}'.format(sender, instance, model))
    CacheService.invalidate_cache_for(instance)



# -*- coding:utf-8 -*-
from setuptools import setup, find_packages

setup(
    name='django_caching',
    version='1.1.0',
    keywords=('django', 'cache', 'caching'),
    description='A easy using django cache manager.',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: MacOS X',
        'Environment :: Plugins',
        'Environment :: Web Environment',
        'Environment :: Win32 (MS Windows)',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: Microsoft :: Windows :: Windows 8',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    license='Apache License',
    install_requires=['django>=1.8'],

    author='NineChapter',
    author_email='daniel48@126.com',
    maintainer='NineChapter',
    maintainer_email='info@ninechapter.com',
    download_url='https://github.com/ninechapter/django_caching',
    url='https://github.com/ninechapter/django_caching',

    packages=find_packages(),
    platforms='any',
)



import uuid

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.cache import caches

from cache.models import Invalidation

cache = caches['test'] if settings.TESTING else caches['cache_manager']


class CacheService(object):
    @classmethod
    def bulk_create_invalidation(cls, cached_objects, key, sql):
        # remove the existing invalidations
        # because this method only be called when cache miss
        # in case of memcached's memory is low, this deletion is
        # necessary, otherwise a lot of legacy invalidations will
        # be created because of quick eviction & cache miss.
        Invalidation.objects.filter(key=key).delete()

        # create new invalidations
        invs = Invalidation.objects.bulk_create([
            Invalidation(
                key=key,
                cached_object=obj,
                sql=sql,
                count=len(cached_objects),
            )
            for obj in cached_objects
        ])
        return invs

    @classmethod
    def get_invalidations(cls, obj):
        content_type = ContentType.objects.get_for_model(obj)
        return Invalidation.objects.filter(
            content_type=content_type,
            object_id=obj.id,
        )

    @classmethod
    def invalidate_cache_for(cls, obj):
        # if obj.id is not set, that means this is a new created
        # object, we need to reset the cache namespace to make
        # some queries like objects.all() work.
        if not obj.id:
            cls._reset_namespace(obj._meta.db_table)
            return

        # TODO: Think about raise condition
        # if new invalidation created when invalidating
        # what will happen?
        #
        # different with bulk_create_invalidation, we don't
        # need to remove the legacy invalidations here
        # because the removal from db may cause efficiency issue
        # why we don't have efficiency problem in bulk_create?
        # because we need *bulk* create anyway, that's already a
        # slow db query.
        invalidations = cls.get_invalidations(obj)
        keys = [i.key for i in invalidations]
        cache.delete_many(keys)
        invalidations.delete()

    # cache namespace

    @classmethod
    def _namespace_key(cls, model):
        return 'namespace::{}'.format(model)

    @classmethod
    def _reset_namespace(cls, model):
        key = cls._namespace_key(model)
        namespace = uuid.uuid4().hex
        # namespace should never expire
        cache.set(key, namespace, timeout=None)
        return namespace

    @classmethod
    def get_namespace(cls, model):
        key = cls._namespace_key(model)
        namespace = cache.get(key)
        if not namespace:
            namespace = cls._reset_namespace(model)
        return namespace



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models.signals import (
    pre_save,
    pre_delete,
    m2m_changed,
)

from cache.listeners import (
    invalidate_model_cache,
    invalidate_m2m_cache,
)


class Invalidation(models.Model):
    """
    Invalidation is for storing the cached object and the related keys
    An db object could be cached for in many queries, for example:

    1. Problem.cached_objects.get(id=1)
    2. Problem.cached_objects.get(unique_name='a-b-plus')
    3. Problem.cached_objects.all()

    we will generate three keys and store three different results in cache.
    once the object is changed, we need to invalidate the three related keys
    in the cache.

    so we can query from the Invalidation table and find the three keys.
    """

    key = models.CharField(max_length=255, help_text='cache key', db_index=True)
    object_id = models.IntegerField()
    content_type = models.ForeignKey(ContentType, null=True)
    cached_object = GenericForeignKey('content_type', 'object_id')

    # for admin
    sql = models.TextField(null=True)
    count = models.IntegerField(null=True)
    created_at = models.DateTimeField(null=True, auto_now_add=True)

    # deprecated
    class_name = models.CharField(max_length=255, null=True)

    class Meta:
        index_together = ['content_type', 'object_id']

    def __str__(self):
        return '{}.{}.{}'.format(self.content_type, self.object_id, self.key)

    def __unicode__(self):
        return u'{}'.format(self.__str__())


pre_save.connect(invalidate_model_cache)
pre_delete.connect(invalidate_model_cache)
m2m_changed.connect(invalidate_m2m_cache)
